import numpy as np
from typing import List, Union, Dict, Tuple
from scipy.optimize import LinearConstraint, minimize
from scipy.stats._distn_infrastructure import rv_frozen

from ..Graph import Graph
from ..Trace import Traces, PseudoTraces
from .BaseWeightEstimator import BaseGLTWEightEstimator

__all__ = ["GLTGridSearchEstimator", "GLTWeightEstimator"]


class GLTWeightEstimator(BaseGLTWEightEstimator):
    """GLTW Weight Estimator for modeling influence in networks.

    Attributes
    ----------
    vertex_2_distrib : Dict[int, rv_frozen]
        Distribution mapping for vertices.
    """

    def _compute_failed_vertex_ll(self, weights: np.ndarray, vertex: int) -> float:
        """Compute the log-likelihood for failed vertices.

        Parameters
        ----------
        weights : np.ndarray
            The weights for the graph edges.
        vertex : int
            The vertex index.

        Returns
        -------
        float
            The log-likelihood for the failed vertex.
        """
        if vertex not in self._failed_vertices_masks:
            return 0.0
        cdf = self.vertex_2_distrib[vertex].cdf
        failed_vertices_indeg = self._failed_vertices_masks[vertex] @ weights
        failure_probs = 1.0 - cdf(failed_vertices_indeg)
        ll = np.log(np.clip(failure_probs, 1e-8, None)).sum()
        return ll

    def _compute_activated_vertex_ll(self, weights: np.ndarray, vertex: int) -> float:
        """Compute the log-likelihood for activated vertices.

        Parameters
        ----------
        weights : np.ndarray
            The weights for the graph edges.
        vertex : int
            The vertex index.

        Returns
        -------
        float
            The log-likelihood for the activated vertex.
        """
        if vertex not in self._vertex_2_active_parent_mask_tm1:
            return 0.0
        cdf = self.vertex_2_distrib[vertex].cdf
        activated_vertices_indeg_tm1 = self._vertex_2_active_parent_mask_tm1[vertex] @ weights
        activated_vertices_indeg_t = self._vertex_2_active_parent_mask_t[vertex] @ weights
        activation_probs = cdf(activated_vertices_indeg_t) - cdf(activated_vertices_indeg_tm1)
        ll = np.log(np.clip(activation_probs, 1e-8, None)).sum()
        return ll

    def _compute_vertex_nll(self, weights: np.ndarray, vertex: int) -> float:
        """Compute the negative log-likelihood for a vertex.

        Parameters
        ----------
        weights : np.ndarray
            The weights for the graph edges.
        vertex : int
            The vertex index.

        Returns
        -------
        float
            The negative log-likelihood for the vertex.
        """
        ll = self._compute_activated_vertex_ll(weights, vertex) + self._compute_failed_vertex_ll(weights, vertex)
        return -ll

    def _compute_total_nll(self, weights: np.ndarray) -> float:
        """Compute the total negative log-likelihood for informative vertices.

        Parameters
        ----------
        weights : np.ndarray
            The weights for the graph edges.

        Returns
        -------
        float
            The total negative log-likelihood.
        """
        return np.sum([
            self._compute_vertex_nll(weights[self.graph.get_parents_mask(vertex)], vertex)
            for vertex in self.informative_vertices
        ])

    def _compute_normalized_vertex_nll(self, weights: np.ndarray, vertex: int) -> float:
        """Compute the normalized negative log-likelihood for a vertex.

        Parameters
        ----------
        weights : np.ndarray
            The weights for the graph edges.
        vertex : int
            The vertex index.

        Returns
        -------
        float
            The normalized negative log-likelihood for the vertex.
        """
        return self._compute_vertex_nll(weights, vertex) / self._vertex_2_num_traces_was_informative[vertex]

    def _make_parent_weight_constraints(self, vertex: int, eps: float = 1e-8) -> LinearConstraint:
        """Create constraints for parent weights of a vertex.

        Parameters
        ----------
        vertex : int
            The vertex index.
        eps : float, optional
            Small value for numerical stability.

        Returns
        -------
        LinearConstraint
            The linear constraints for optimization.
        """
        indeg_max = self.vertex_2_distrib[vertex].support()[1]
        num_parents = self.graph.get_indegree(vertex, weighted=False)

        mat_constrain = np.vstack([np.eye(num_parents), np.ones(num_parents)])
        lb = eps * np.ones(num_parents + 1)
        ub = (indeg_max - eps) * np.ones(num_parents + 1)
        return LinearConstraint(A=mat_constrain, lb=lb, ub=ub)

    def _precompute_all_constraints(self) -> None:
        """Precompute weight constraints for all informative vertices."""
        self._weight_constraints = {
            vertex: self._make_parent_weight_constraints(vertex)
            for vertex in self.informative_vertices
        }

    def _optimize_vertex_parent_params(self, vertex: int, optimization_kwargs: Dict = None) -> Tuple[np.ndarray, rv_frozen]:
        """Optimize parameters for a specific vertex.

        Parameters
        ----------
        vertex : int
            The vertex to optimize.
        optimization_kwargs : Dict, optional
            Additional optimization parameters.

        Returns
        -------
        Tuple[np.ndarray, rv_frozen]
            Optimized weights and the vertex distribution.
        """
        optimizer_output = minimize(
            lambda weights: self._compute_normalized_vertex_nll(weights, vertex=vertex),
            x0=self.weights_[self.graph.get_parents_mask(vertex)],
            method='SLSQP',
            constraints=self._weight_constraints[vertex],
            options=optimization_kwargs
        )
        return optimizer_output.x, self.vertex_2_distrib[vertex]

    def _set_informative_vertices_parent_params(self, informative_vertices_params: List[Tuple[np.ndarray, rv_frozen]]) -> None:
        """Set the optimized parameters for informative vertices.

        Parameters
        ----------
        informative_vertices_params : List[Tuple[np.ndarray, rv_frozen]]
            List of optimized parameters for each informative vertex.
        """
        for vertex, (parent_weights, vertex_distrib) in zip(self.informative_vertices, informative_vertices_params):
            self.weights_[self.graph.get_parents_mask(vertex)] = parent_weights
            self.vertex_2_distrib[vertex] = vertex_distrib

    def _pre_fit(self, traces: Union[Traces, PseudoTraces], 
                 init_weights: Union[List[float], np.array] = None, 
                 masks_path: str = None) -> None:
        """Prepare for fitting by preprocessing traces and initializing constraints.

        Parameters
        ----------
        traces : Union[Traces, PseudoTraces]
            The traces to analyze.
        init_weights : Optional[List[float]]
            Initial weights to set.
        masks_path : Optional[str]
            Path to load masks from.
        """
        self._preprocess_traces(traces, masks_path=masks_path)
        self._precompute_all_constraints()
        self._init_weights(init_weights)


class GLTGridSearchEstimator(GLTWeightEstimator):
    """GLT Grid Search Estimator for hyperparameter tuning.

    Attributes
    ----------
    vertex_2_distrib_grid : Dict[int, List[rv_frozen]]
        Distribution grid for each vertex.
    """

    def __init__(self, graph: Graph,
                 distribs_grid: Union[Dict[int, List[rv_frozen]], List[rv_frozen]],
                 n_jobs: int = 1) -> None:
        """Initialize the GLT Grid Search Estimator.

        Parameters
        ----------
        graph : Graph
            The graph structure.
        distribs_grid : Union[Dict[int, List[rv_frozen]], List[rv_frozen]]
            The grid of distributions for each vertex.
        n_jobs : int, optional
            Number of jobs for parallel processing.
        """
        super().__init__(graph=graph, vertex_2_distrib=None, n_jobs=n_jobs)
        self._validate_grid(distribs_grid)

    def _validate_grid(self, distribs_grid: Union[Dict[int, List[rv_frozen]], List[rv_frozen]]) -> None:
        """Validate the distribution grid.

        Parameters
        ----------
        distribs_grid : Union[Dict[int, List[rv_frozen]], List[rv_frozen]]
            The grid of distributions to validate.

        Raises
        ------
        AssertionError
            If the distribution grid does not match the graph vertices.
        """
        if isinstance(distribs_grid, (List, Tuple)):
            self.vertex_2_distrib_grid = {vertex: distribs_grid for vertex in self.graph.get_vertices()}
        elif isinstance(distribs_grid, Dict):
            assert set(distribs_grid.keys()) == self.graph.get_vertices(), \
                "If `distribs_grid` is a dict, its keys should be the vertices of the graph."
            self.vertex_2_distrib_grid = distribs_grid
        else:
            raise NotImplementedError("`distribs_grid` should be either a list, tuple, or a dict.")

        assert all([
            all([isinstance(distrib, rv_frozen) for distrib in vertex_distribs])
            for vertex_distribs in self.vertex_2_distrib_grid.values()
        ]), "All elements of the grid should be `rv_frozen`."

    def _optimize_vertex_parent_params(self, vertex: int, optimization_kwargs: Dict = None) -> Tuple[np.ndarray, rv_frozen]:
        """Optimize parameters for a specific vertex using grid search.

        Parameters
        ----------
        vertex : int
            The vertex to optimize.
        optimization_kwargs : Dict, optional
            Additional optimization parameters.

        Returns
        -------
        Tuple[np.ndarray, rv_frozen]
            Best weights and the best distribution for the vertex.
        """
        best_weights = None
        best_nll = np.inf
        best_distrib = None

        for distrib in self.vertex_2_distrib_grid[vertex]:
            self.vertex_2_distrib[vertex] = distrib
            optimizer_output = minimize(
                lambda weights: self._compute_normalized_vertex_nll(weights, vertex=vertex),
                x0=self.weights_[self.graph.get_parents_mask(vertex)],
                method='SLSQP',
                constraints=self._weight_constraints[vertex],
                options=optimization_kwargs
            )
            nll = optimizer_output.fun
            if nll < best_nll:
                best_weights = optimizer_output.x
                best_distrib = distrib
                best_nll = nll

        return best_weights, best_distrib



import pandas as pd
import numpy as np
from typing import List, Tuple, Union, Dict
import networkx as nx

__all__ = ["Graph"]


class Graph(nx.DiGraph):
    def __init__(self, edge_list: List[Tuple[int, int]],
                 directed: bool = True,
                 weights: Union[np.array, List[float]] = None):
        """
        Initialize a directed or undirected graph from an edge list.

        Parameters
        ----------
        edge_list : List[Tuple[int, int]]
            List of tuples (source, sink) representing edges of the graph.
        directed : bool, optional
            Flag to indicate whether the graph is directed (default is True).
        weights : Union[None, List[float]], optional
            Optional weights for each edge. If None, all edges are assigned a weight of 1.
        """
        self.directed = directed
        self.edge_array = np.array(edge_list, dtype=int)

        if not self.directed:
            reverse_edge_array = self.edge_array[:, [1, 0]]
            self.edge_array = np.concatenate([self.edge_array, reverse_edge_array])

        super().__init__(self.edge_array.tolist())

        if weights is not None:
            assert len(weights) == len(edge_list), "number of edges is different from number of weights"
        else:
            weights = np.ones(len(edge_list))
        self.weights = np.array(weights) if directed else np.hstack([weights] * 2)
        self.edge_2_index = {tuple(edge): idx for idx, edge in enumerate(self.edge_array)}

    def rename_graph_vertices(self, old_2_new_vertex_name_dict: Union[None, Dict] = None) -> "Graph":
        """
        Rename graph vertices according to the provided mapping or re-index them.

        Parameters
        ----------
        old_2_new_vertex_name_dict : Union[None, Dict], optional
            Mapping of old vertex names to new names. 
            If None, re-indexes vertices to [0, 1, ..., |V|-1].

        Returns
        -------
        Graph
            The updated graph instance with renamed vertices.
        """
        if old_2_new_vertex_name_dict is None:
            old_vertex_names = sorted(list(self.get_vertices()))
            old_2_new_vertex_name_dict = {name: idx for idx, name in enumerate(old_vertex_names)}

        self.edge_array[:, 0] = list(map(lambda v: old_2_new_vertex_name_dict[v], self.edge_array[:, 0]))
        self.edge_array[:, 1] = list(map(lambda v: old_2_new_vertex_name_dict[v], self.edge_array[:, 1]))
        return self

    def set_weights(self, weights: Union[str, np.array]) -> "Graph":
        """
        Set weights for the edges in the graph.

        Parameters
        ----------
        weights : Union[str, np.array]
            Array of weights for the edges.

        Returns
        -------
        Graph
            The current graph instance with updated weights.
        
        Raises
        ------
        AssertionError
            If the number of weights does not match the number of edges.
        """
        assert len(weights) == self.count_edges(), "number of weights different from the number of edges"
        self.weights = np.array(weights)
        return self

    def get_edges(self, as_array: bool = False) -> Union[np.array, List[Tuple[int, int]]]:
        """
        Retrieve the edges of the graph.

        Parameters
        ----------
        as_array : bool, optional
            If True, returns edges as a numpy array. If False, returns as a list of tuples (default is False).

        Returns
        -------
        Union[np.array, List[Tuple[int, int]]]
            The edges of the graph in the specified format.
        """
        if as_array:
            return self.edge_array
        return [tuple(edge) for edge in self.edge_array]

    def get_vertices(self) -> set:
        """
        Get the set of all vertices in the graph.

        Returns
        -------
        set
            A set containing all vertex identifiers.
        """
        return set(self.nodes)
    
    def get_sources(self) -> set: 
        """
        Get the set of source vertices in the graph.

        Returns
        -------
        set
            A set containing all source vertices.
        """
        return set(self.edge_array[:, 0])
    
    def get_sinks(self) -> set: 
        """
        Get the set of sink vertices in the graph.

        Returns
        -------
        set
            A set containing all sink vertices.
        """
        return set(self.edge_array[:, 1])

    def count_edges(self) -> int:
        """
        Count the number of edges in the graph.

        Returns
        -------
        int
            The total number of edges.
        """
        return len(self.edge_array)

    def count_vertices(self) -> int:
        """
        Count the number of vertices in the graph.

        Returns
        -------
        int
            The total number of vertices.
        """
        return len(self.get_vertices())

    def get_children_mask(self, vertex: int) -> np.array:
        """
        Create a mask for edges originating from a specified vertex.

        Parameters
        ----------
        vertex : int
            The vertex for which to find children.

        Returns
        -------
        np.array
            A boolean array mask indicating edges originating from the vertex.
        """
        return self.edge_array[:, 0] == vertex

    def get_children(self, vertex: int, out_type: Union[set, list, np.array] = set) -> Union[set, list, np.array]:
        """
        Get the children (outgoing edges) of a specified vertex.

        Parameters
        ----------
        vertex : int
            The vertex for which to find children.
        out_type : Union[set, list, np.array], optional
            The type of output to return (default is set).

        Returns
        -------
        Union[set, list, np.array]
            The children of the vertex in the specified output type.
        """
        return out_type(self.edge_array[:, 1][self.get_children_mask(vertex)])

    def get_parents_mask(self, vertex: int) -> np.array:
        """
        Create a mask for edges leading to a specified vertex.

        Parameters
        ----------
        vertex : int
            The vertex for which to find parents.

        Returns
        -------
        np.array
            A boolean array mask indicating edges leading to the vertex.
        """
        return self.edge_array[:, 1] == vertex

    def get_parents(self, vertex: int, out_type: Union[set, list, np.array] = set) -> Union[set, list, np.array]:
        """
        Get the parents (incoming edges) of a specified vertex.

        Parameters
        ----------
        vertex : int
            The vertex for which to find parents.
        out_type : Union[set, list, np.array], optional
            The type of output to return (default is set).

        Returns
        -------
        Union[set, list, np.array]
            The parents of the vertex in the specified output type.
        """
        return out_type(self.edge_array[:, 0][self.get_parents_mask(vertex)])

    def get_vertex_2_indegree_dict(self, weighted: bool = False) -> Dict[int, Union[float, int]]:
        """
        Get a dictionary mapping vertices to their indegrees.

        Parameters
        ----------
        weighted : bool, optional
            If True, returns weighted indegrees; otherwise, returns simple counts (default is False).

        Returns
        -------
        Dict[int, Union[float, int]]
            A dictionary where keys are vertex identifiers and values are their indegrees.
        """
        return {vertex: self.get_indegree(vertex, weighted=weighted) for vertex in self.get_vertices()}

    def get_outdegree(self, vertex: int, weighted: bool = False) -> Union[float, int]:
        """
        Get the outdegree of a specified vertex.

        Parameters
        ----------
        vertex : int
            The vertex for which to find the outdegree.
        weighted : bool, optional
            If True, returns the weighted outdegree; otherwise, returns simple counts (default is False).

        Returns
        -------
        Union[float, int]
            The outdegree of the vertex.
        """
        children_weights = self.weights[self.edge_array[:, 0] == vertex]
        return np.sum(children_weights) if weighted else len(children_weights)

    def get_all_outdegrees(self, weighted: bool = False) -> np.array:
        """
        Get outdegrees for all vertices in the graph.

        Parameters
        ----------
        weighted : bool, optional
            If True, returns weighted outdegrees; otherwise, returns simple counts (default is False).

        Returns
        -------
        np.array
            An array of outdegrees for all vertices.
        """
        return np.array([self.get_outdegree(v, weighted=weighted) for v in self.get_vertices()])

    def get_avg_outdegree(self, weighted: bool = False) -> Union[float, int]:
        """
        Calculate the average outdegree of all vertices.

        Parameters
        ----------
        weighted : bool, optional
            If True, uses weighted outdegrees; otherwise, uses simple counts (default is False).

        Returns
        -------
        Union[float, int]
            The average outdegree of the vertices.
        """
        return np.mean(self.get_all_outdegrees(weighted=weighted))

    def get_max_outdegree(self, weighted: bool = False) -> Union[float, int]:
        """
        Get the maximum outdegree among all vertices.

        Parameters
        ----------
        weighted : bool, optional
            If True, considers weighted outdegrees; otherwise, uses simple counts (default is False).

        Returns
        -------
        Union[float, int]
            The maximum outdegree.
        """
        return np.max(self.get_all_outdegrees(weighted))

    def get_indegree(self, vertex: int, weighted: bool = False) -> Union[float, int]:
        """
        Get the indegree of a specified vertex.

        Parameters
        ----------
        vertex : int
            The vertex for which to find the indegree.
        weighted : bool, optional
            If True, returns the weighted indegree; otherwise, returns simple counts (default is False).

        Returns
        -------
        Union[float, int]
            The indegree of the vertex.
        """
        parent_weights = self.weights[self.edge_array[:, 1] == vertex]
        return np.sum(parent_weights) if weighted else len(parent_weights)
    
    def get_indegrees_dict(self, weighted: bool = False) -> Dict[int, Union[int, float]]:
        """
        Get a mapping from vertex to its indegree for all vertices in the graph.

        Parameters
        ----------
        weighted : bool, optional
            If True, returns weighted indegrees; otherwise, returns simple counts (default is False).

        Returns
        -------
        np.array
            An array of indegrees for all vertices.
        """
        return {v: self.get_indegree(v, weighted=weighted) for v in self.get_vertices()}

    def get_all_indegrees(self, weighted: bool = False) -> np.array:
        """
        Get indegrees for all vertices in the graph.

        Parameters
        ----------
        weighted : bool, optional
            If True, returns weighted indegrees; otherwise, returns simple counts (default is False).

        Returns
        -------
        np.array
            An array of indegrees for all vertices.
        """
        indegrees_dict = self.get_indegrees_dict(weighted=weighted)
        return np.array(list(indegrees_dict.values()))

    def get_max_indegree(self, weighted: bool = False) -> Union[int, float]:
        """
        Get the maximum indegree among all vertices.

        Parameters
        ----------
        weighted : bool, optional
            If True, considers weighted indegrees; otherwise, uses simple counts (default is False).

        Returns
        -------
        Union[int, float]
            The maximum indegree.
        """
        return np.max(self.get_all_indegrees(weighted))

    def get_avg_indegree(self, weighted: bool = False) -> float:
        """
        Calculate the average indegree of all vertices.

        Parameters
        ----------
        weighted : bool, optional
            If True, uses weighted indegrees; otherwise, uses simple counts (default is False).

        Returns
        -------
        float
            The average indegree of the vertices.
        """
        return np.mean(self.get_all_indegrees(weighted))

    def get_edge_index(self, edge: Tuple[int, int]) -> int:
        """
        Get the index of a specified edge.

        Parameters
        ----------
        edge : Tuple[int, int]
            The edge for which to find the index.

        Returns
        -------
        int
            The index of the edge in the graph.
        """
        return self.edge_2_index[tuple(edge)]

    def get_edge_weight(self, edge: Tuple[int, int]) -> float:
        """
        Get the weight of a specified edge.

        Parameters
        ----------
        edge : Tuple[int, int]
            The edge for which to retrieve the weight.

        Returns
        -------
        float
            The weight of the edge.
        """
        return self.weights[self.get_edge_index(edge)]

    def get_edges_mask_from_set_to_vertex(self, vertex_set: set, vertex: int) -> np.array:
        """
        Create a mask for edges leading from a set of vertices to a specified vertex.

        Parameters
        ----------
        vertex_set : set
            A set of source vertices.
        vertex : int
            The target vertex.

        Returns
        -------
        np.array
            A boolean array mask indicating which edges lead from vertices in vertex_set to the specified vertex.
        """
        parent_edges_mask = self.get_parents_mask(vertex)
        parent_vertices = self.edge_array[:, 0][parent_edges_mask]
        parent_edges_mask[parent_edges_mask] = [(parent in vertex_set) for parent in parent_vertices]
        return parent_edges_mask

    def get_adj_matrix(self) -> np.array:
        """
        Generate the adjacency matrix of the graph.

        Returns
        -------
        np.array
            The adjacency matrix as a numpy array.
        """
        vertices = list(self.get_vertices())
        n_vertex = len(vertices)
        adj_matrix = pd.DataFrame(np.zeros((n_vertex, n_vertex)), columns=vertices, index=vertices)
        for (v1, v2), weight in zip(self.edge_array, self.weights):
            adj_matrix.at[v1, v2] = weight

        return adj_matrix

    def is_edge_in_graph(self, edge: Tuple[int, int]) -> bool:
        """
        Check if a specified edge exists in the graph.

        Parameters
        ----------
        edge : Tuple[int, int]
            The edge to check.

        Returns
        -------
        bool
            True if the edge exists in the graph, False otherwise.
        """
        source, sink = edge
        return sink in self.get_children(source)
